﻿using System;
using System.Runtime.CompilerServices;
using System.Threading.Tasks;

namespace ConsoleApp3
{
    // 
    // ConsoleApp1 예제에서 CallAsync, GetFileContents 메서드의 async를 해제
    //
    class Program
    {
        static void Main(string[] args)
        {
            Program pg = new Program();
            Task task = pg.CallAsync();

            task.GetAwaiter().GetResult();
        }

        private Task CallAsync()
        {
            CallAsync_StateMachine stateMachine = new CallAsync_StateMachine
            {
                _this = this,
                _builder = AsyncTaskMethodBuilder.Create(),
                _state = -1,
            };

            stateMachine._builder.Start(ref stateMachine);
            return stateMachine._builder.Task;
        }

        class CallAsync_StateMachine : IAsyncStateMachine
        {
            public int _state; // 1
            public AsyncTaskMethodBuilder _builder;
            public Program _this; // 4

            string _result; // 3

            // 반환값
            string _text;

            // await 호출 전 코드의 지역 변수
            string _title;

            TaskAwaiter<string> _awaiter;

            void IAsyncStateMachine.MoveNext()
            {
                int num = this._state;

                try
                {
                    TaskAwaiter<string> awaiter;
                    switch (num)
                    {
                        case 0:
                            awaiter = this._awaiter;
                            this._awaiter = new TaskAwaiter<string>();
                            this._state = num = -1;
                            break;

                        default:

                            this._title = DateTime.Now.ToString();

                            awaiter = _this.GetFileContents().GetAwaiter();
                            if (awaiter.IsCompleted == false)
                            {
                                this._state = num = 0;
                                this._awaiter = awaiter;
                                CallAsync_StateMachine stateMachine = this;
                                this._builder.AwaitUnsafeOnCompleted(ref awaiter, ref stateMachine);
                                return;
                            }
                            break;
                    }

                    this._result = awaiter.GetResult();
                    this._text = this._result;
                    this._result = null;
                    Console.WriteLine(this._title + ": " + this._text);
                }
                catch (Exception e)
                {
                    this._state = -2;
                    this._builder.SetException(e);
                    return;
                }

                this._state = -2;
                this._builder.SetResult();
            }

            void IAsyncStateMachine.SetStateMachine(IAsyncStateMachine stateMachine)
            {
            }
        }

        private Task<string> GetFileContents()
        {
            GetFileContents_StateMachine stateMachine = new GetFileContents_StateMachine
            {
                _this = this,
                _builder = AsyncTaskMethodBuilder<string>.Create(),
                _state = -1,
            };

            stateMachine._builder.Start(ref stateMachine);
            return stateMachine._builder.Task as Task<string>;
        }

        class GetFileContents_StateMachine : IAsyncStateMachine
        {
            public int _state; // 1
            public AsyncTaskMethodBuilder<string> _builder;
            public Program _this; // 4

            string _result; // 3

            // 반환값
            string _urlContents;
            Task<string> _getStringTask;

            TaskAwaiter<string> _awaiter;

            void IAsyncStateMachine.MoveNext()
            {
                string str;
                int num = this._state;

                try
                {
                    TaskAwaiter<string> awaiter;
                    switch (num)
                    {
                        case 0:
                            awaiter = this._awaiter;
                            this._awaiter = new TaskAwaiter<string>();
                            this._state = num = -1;
                            break;

                        default:
                            this._getStringTask = new TaskFactory().StartNew(() => { return "test"; });
                            awaiter = this._getStringTask.GetAwaiter();
                            if (awaiter.IsCompleted == false)
                            {
                                this._state = num = 0;
                                this._awaiter = awaiter;
                                GetFileContents_StateMachine stateMachine = this;
                                this._builder.AwaitUnsafeOnCompleted(ref awaiter, ref stateMachine);
                                return;
                            }
                            break;
                    }

                    this._result = awaiter.GetResult();
                    this._urlContents = this._result;
                    this._result = null;

                    str = this._urlContents;
                }
                catch (Exception e)
                {
                    this._state = -2;
                    this._builder.SetException(e);
                    return;
                }

                this._state = -2;
                this._builder.SetResult(str);
            }

            void IAsyncStateMachine.SetStateMachine(IAsyncStateMachine stateMachine)
            {
                Console.WriteLine((stateMachine as GetFileContents_StateMachine)._state);
            }
        }

    }

}
